/*
 * Ukazkovy program pro 3. cv. ZPG
 * 
 * Jiri Skala, 2006
 * 
 */
using System;
using System.Drawing;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace zpg_cv03
{
    class ZpgD3D : Form
    {
        /// <summary>
        /// Direct3D device.
        /// </summary>
        private Device device = null;

        /// <summary>
        /// Direct3D vertx buffer.
        /// </summary>
        private VertexBuffer vertexBuff = null;

        /// <summary>
        /// Yaw - rotation about vertical axis. That is looking around.
        /// </summary>
        private float yaw = 0;

        /// <summary>
        /// Pitch - rotation about lateral axis. That is looking up and down.
        /// </summary>
        private float pitch = 0;

        /// <summary>
        /// For measuring time intervals.
        /// </summary>
        private long ticks = 0;

        /// <summary>
        /// Angle of rotation for the second cube.
        /// </summary>
        private float angle1 = 0;

        /// <summary>
        /// Angle of rotation for the third cube.
        /// </summary>
        private float angle2 = 0;

        /// <summary>
        /// Controls mouse sensitivity.
        /// Higher number means lower sensitivity.
        /// </summary>
        private const float MouseLookFactor = 500.0f;



        /// <summary>
        /// Constructor.
        /// </summary>
        public ZpgD3D()
        {
            // set window properties
            this.Size = new Size(640, 480);
            this.Text = "Direct3D example";

            // register "mouse moved" event handler
            this.MouseMove += new MouseEventHandler(ZpgD3DMouseMove);
            // register "key down" event handler
            this.KeyDown += new KeyEventHandler(ZpgD3DKeyDown);
        }

        /// <summary>
        /// "Key down" event handler.
        /// </summary>
        void ZpgD3DKeyDown(object sender, KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.Escape:
                    // close application
                    this.Close();
                    break;
            }
        }

        /// <summary>
        /// "Mouse moved" event handler.
        /// </summary>
        void ZpgD3DMouseMove(object sender, MouseEventArgs e)
        {
            // get window size and compute it's center
            Size s = this.ClientSize;
            int cx = s.Width / 2;
            int cy = s.Height / 2;

            // check whether we should process this event
            if (e.X == cx && e.Y == cy)
                // in fact nothing happened
                // this event was generated when we moved the mouse by setting Cursor.Position
                return;

            // update "looking" angles
            yaw += (float)(cx - e.X) / MouseLookFactor;
            pitch += (float)(cy - e.Y) / MouseLookFactor;

            // limit angles to some reasonable range
            if (yaw > 2 * Math.PI)
                yaw -= 2 * (float)Math.PI;
            if (pitch > Math.PI / 2)
                pitch = (float)Math.PI / 2;
            else if (pitch < -Math.PI / 2)
                pitch = -(float)Math.PI / 2;

            // we have center relative to current window
            // compute center position relative to desktop
            Point center = PointToScreen(new Point(cx, cy));
            // position the (invisible) cursor in the middle of the window
            Cursor.Position = center;
        }

        /// <summary>
        /// Direct3D initialization.
        /// </summary>
        /// <returns></returns>
        private bool InitD3D()
        {
            try
            {
                // prepare presentation parameters
                PresentParameters presParams = new PresentParameters();
                presParams.Windowed = true;
                presParams.SwapEffect = SwapEffect.Discard;
                presParams.BackBufferFormat = Format.Unknown;
                presParams.EnableAutoDepthStencil = false;

                // determine graphic card capabilities
                Caps caps = Manager.GetDeviceCaps(0, DeviceType.Hardware);
                CreateFlags flags;
                if (caps.DeviceCaps.SupportsHardwareTransformAndLight)
                    flags = CreateFlags.HardwareVertexProcessing;
                else
                    flags = CreateFlags.SoftwareVertexProcessing;

                // create D3D device
                device = new Device(0, DeviceType.Hardware, this, flags, presParams);

                // create vertex buffer
                vertexBuff = new VertexBuffer(typeof(CustomVertex.PositionOnly), 10, device,
                    Usage.WriteOnly, CustomVertex.PositionOnly.Format, Pool.Managed);
                // fill the buffer with vertices
                CustomVertex.PositionOnly[] verts
                    = (CustomVertex.PositionOnly[])vertexBuff.Lock(0, LockFlags.None);
                verts[0] = new CustomVertex.PositionOnly(-1, -1, -1);
                verts[1] = new CustomVertex.PositionOnly( 1, -1, -1);
                verts[2] = new CustomVertex.PositionOnly(-1,  1, -1);
                verts[3] = new CustomVertex.PositionOnly( 1,  1, -1);
                verts[4] = new CustomVertex.PositionOnly(-1,  1,  1);
                verts[5] = new CustomVertex.PositionOnly( 1,  1,  1);
                verts[6] = new CustomVertex.PositionOnly(-1, -1,  1);
                verts[7] = new CustomVertex.PositionOnly( 1, -1,  1);
                verts[8] = new CustomVertex.PositionOnly(-1, -1, -1);
                verts[9] = new CustomVertex.PositionOnly( 1, -1, -1);
                vertexBuff.Unlock();

                return true;
            }
            catch (DirectXException)
            {
                // something is wrong
                MessageBox.Show("Direct3D initialization failed.", "Error",
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
                return false;
            }
        }

        /// <summary>
        /// Render the scene.
        /// </summary>
        private void Render()
        {
            // some local constants
            const float bigCubeScale = 0.5f;
            const float smallCubeScale = 0.3f;
            const float secondCubeMove = 3.0f;
            const float thirdCubeMove = 1.5f;
            const float secondCubeRotFactor = 500f;
            const float thirdCubeRotFactor = -180f;

            if (device == null)
                return;

            // measure elapsed time
            long newTicks = Environment.TickCount;
            float timeElapsed = newTicks - ticks;
            ticks = newTicks;

            // update angles according to elapsed time
            angle1 += timeElapsed / secondCubeRotFactor;
            angle2 += timeElapsed / thirdCubeRotFactor;

            // limit angles
            if (angle1 > 2 * Math.PI)
                angle1 -= 2 * (float)Math.PI;
            if (angle2 > 2 * Math.PI)
                angle2 -= 2 * (float)Math.PI;

            // set rendering properties
            device.RenderState.Lighting = false;
            device.RenderState.CullMode = Cull.None;
            device.RenderState.FillMode = FillMode.WireFrame;

            // set up perspective projection
            float aspect = (float)this.ClientSize.Width / (float)this.ClientSize.Height;
            device.Transform.Projection = Matrix.PerspectiveFovLH((float)Math.PI / 4, aspect, 0.1f, 30.0f);

            // move camera so that we can see something
            device.Transform.View = Matrix.Identity;
            device.Transform.View *= Matrix.RotationX(pitch);
            device.Transform.View *= Matrix.RotationY(yaw);
            device.Transform.View *= Matrix.Translation(0, 0, 10);

            // clear scene
            device.Clear(ClearFlags.Target, Color.DarkBlue, 1, 0);
            // start rendering
            device.BeginScene();

            // prepare for rendering from the vertex buffer
            device.SetStreamSource(0, vertexBuff, 0);
            device.VertexFormat = CustomVertex.PositionOnly.Format;

            // compute transformation for the first cube
            device.Transform.World = Matrix.Identity;
            device.Transform.World *= Matrix.Scaling(bigCubeScale, bigCubeScale, bigCubeScale);
            // draw the first cube
            device.DrawPrimitives(PrimitiveType.TriangleStrip, 0, 8);

            // append transformation for the second cube
            Matrix secondCubeTransform = Matrix.Translation(secondCubeMove, 0, 0);
            secondCubeTransform *= Matrix.RotationY(angle1);
            device.Transform.World *= secondCubeTransform;
            // draw second cube
            device.DrawPrimitives(PrimitiveType.TriangleStrip, 0, 8);

            // compute transformation for the third cube
            device.Transform.World = Matrix.Identity;
            device.Transform.World *= Matrix.Scaling(smallCubeScale, smallCubeScale, smallCubeScale);
            device.Transform.World *= Matrix.Translation(thirdCubeMove, 0, 0);
            device.Transform.World *= Matrix.RotationY(angle2);
            device.Transform.World *= secondCubeTransform;
            // draw the third cube
            device.DrawPrimitives(PrimitiveType.TriangleStrip, 0, 8);

            // rendering done
            device.EndScene();
            // show result
            device.Present();
        }

        /// <summary>
        /// Main "game" loop.
        /// </summary>
        private void Run()
        {
            // hide mouse cursor
            Cursor.Hide();

            // get window size and compute it's center
            Size s = this.ClientSize;
            int cx = s.Width / 2;
            int cy = s.Height / 2;
            // we have center relative to current window
            // compute center position relative to desktop
            Point center = PointToScreen(new Point(cx, cy));
            // position the (invisible) cursor in the middle of the window
            Cursor.Position = center;

            // initialize tick counter
            ticks = Environment.TickCount;

            // enter "game loop"
            while (this.Created)
            {
                Render();
                Application.DoEvents(); // should be the last in this loop
            }

            // show mouse cursor
            Cursor.Show();

            // dispose Direct3D resources
            if (vertexBuff != null)
                vertexBuff.Dispose();
            if (device != null)
                device.Dispose();
        }

        /// <summary>
        /// Application entry point.
        /// </summary>
        static void Main(string[] args)
        {
            // create application window
            ZpgD3D form = new ZpgD3D();
            // make sure it's visible
            form.Show();

            // initialize Direct3D
            if (form.InitD3D())
            {
                form.Run();
            }
        }
    }
}
